



python -c "import cartopy.crs as ccrs; import matplotlib.pyplot as plt; ax = plt.axes(projection=ccrs.PlateCarree())"
IF %ERRORLEVEL% NEQ 0 exit /B 1
pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest -ra -n 4 --pyargs cartopy
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
